import logging
import multiprocessing
import threading
import time
from datetime import datetime
from typing import Any, Optional

from wiederverwendbar.task_manger.task import Task
from wiederverwendbar.task_manger.trigger import Trigger
from wiederverwendbar.singleton import Singleton

LOGGER = logging.getLogger(__name__)


class Manager:
    lock = threading.Lock()

    def __init__(self,
                 name: Optional[str] = None,
                 worker_count: Optional[int] = None,
                 daemon: bool = False,
                 keep_done_tasks: bool = False,
                 loop_delay: Optional[float] = None,
                 logger: Optional[logging.Logger] = None):
        if name is None:
            name = self.__class__.__name__
        self.name = name
        self._workers: list[threading.Thread] = []
        self._tasks: list[Task] = []
        self._stopped: bool = False
        self._creation_time: datetime = datetime.now()
        self._keep_done_tasks = keep_done_tasks
        self.logger = logger or LOGGER

        # create workers
        if worker_count is None:
            worker_count = multiprocessing.cpu_count()
            if worker_count - 2 < 1:
                worker_count = 1
            if worker_count > 4:
                worker_count = 4
        for i in range(worker_count):
            worker = threading.Thread(name=f"{self.name}.Worker{i}", target=self.loop, daemon=daemon)
            self._workers.append(worker)

        # set loop delay
        if loop_delay is None:
            if self.worker_count > 1:
                loop_delay = 0.001
        self._loop_delay = loop_delay

    def __str__(self):
        return f"{self.__class__.__name__}({self.name})"

    def __del__(self):
        if not self.stopped:
            self.stop()

    @property
    def worker_count(self):
        """
        Number of workers.

        :return: int
        """

        return len(self._workers)

    @property
    def stopped(self):
        """
        Manager stopped flag.

        :return: bool
        """

        with self.lock:
            stopped = self._stopped
        return stopped

    @property
    def creation_time(self) -> datetime:
        """
        Manager creation time.

        :return: datetime
        """

        with self.lock:
            creation_time = self._creation_time
        return creation_time

    def start(self) -> None:
        """
        Start manager.

        :return: None
        """

        self.logger.debug(f"{self}: Starting manager ...")

        # start workers
        for worker in self._workers:
            self.logger.debug(f"{self}: Starting worker '{worker.name}' ...")
            worker.start()

        self.logger.debug(f"{self}: Manager started.")

    def stop(self) -> None:
        """
        Stop manager.

        :return: None
        """

        self.logger.debug(f"{self}: Stopping manager ...")

        # set stopped flag
        with self.lock:
            self._stopped = True

        # wait for workers to finish
        for worker in self._workers:
            if worker.is_alive():
                self.logger.debug(f"{self}: Waiting for worker '{worker.name}' to finish ...")
                worker.join()

        self.logger.debug(f"{self}: Manager stopped.")

    def loop(self, stay_in_loop: Optional[bool] = None) -> None:
        """
        Manager loop. All workers run this loop. If worker_count is 0, you can run this loop manually.

        :param stay_in_loop: Stay in loop flag. If False, loop will break after the first task is run.
        :return: None
        """

        if stay_in_loop is None:
            with self.lock:
                stay_in_loop = bool(self._loop_delay)
        while not self.stopped:
            now = datetime.now()
            current_task = None
            with self.lock:
                for i, task in enumerate(self._tasks):
                    if task.next_run is None:
                        continue
                    if task.next_run > now:
                        continue
                    current_task = self._tasks.pop(i)
                    break
            if current_task is None:
                with self.lock:
                    loop_delay = self._loop_delay
                if loop_delay:
                    time.sleep(loop_delay)
                continue

            self.logger.debug(f"{self}: Running task '{current_task}' ...")

            with self.lock:
                if current_task.time_measurement_before_run:
                    current_task.set_last_run()
                    current_task.set_next_run()

            # run task
            current_task.payload()

            self.logger.debug(f"{self}: Task '{current_task}' successfully run.")

            with self.lock:
                if not current_task.time_measurement_before_run:
                    current_task.set_last_run()
                    current_task.set_next_run()
                if not current_task.is_done:
                    self._tasks.append(current_task)
                else:
                    self.logger.debug(f"{self}: Task '{current_task}' is done.")
                    if self._keep_done_tasks:
                        self._tasks.append(current_task)

            if not stay_in_loop:
                break

    def add_task(self, task: Task):
        """
        Add task to manager.

        :param task:
        :return:
        """

        task.init(self)
        with self.lock:
            self._tasks.append(task)
        self.logger.debug(f"{self}: Task '{task}' added.")

    def remove_task(self, task: Task):
        """
        Remove task from manager.

        :param task:
        :return:
        """

        with self.lock:
            self._tasks.remove(task)
        self.logger.debug(f"{self}: Task '{task}' removed.")

    def task(self,
             name: Optional[str] = None,
             trigger: Optional[Trigger] = None,
             time_measurement_before_run: bool = True,
             return_func: bool = True,
             *args,
             **kwargs) -> Any:
        """
        Task decorator.

        :param name: The name of the task.
        :param trigger: The trigger of the task.
        :param time_measurement_before_run: Time measurement before run flag.
        :param return_func: Return function flag. If True, the function will be returned instead of the task.
        :param args: Args for the task payload.
        :param kwargs: Kwargs for the task payload.
        :return: Task or function
        """

        def decorator(func):
            task = Task(name=name,
                        manager=self,
                        trigger=trigger,
                        time_measurement_before_run=time_measurement_before_run,
                        payload=func,
                        auto_add=True,
                        *args,
                        **kwargs)
            return task if not return_func else func

        return decorator


class ManagerSingleton(Manager, metaclass=Singleton):
    ...



from wiederverwendbar.pydantic.security.hashed_password_model import HashedPasswordModel
from wiederverwendbar.functions.security.hash_password import hash_password


def verify_password(hashed_password_model: HashedPasswordModel, verifying_password: str) -> bool:
    hashed_verify_password = hash_password(
        password=verifying_password,
        encoding=hashed_password_model.encoding,
        hash_function=hashed_password_model.hash_function,
        interactions=hashed_password_model.interactions,
        key_length=hashed_password_model.key_length,
        salt=hashed_password_model.salt
    )
    result = hashed_password_model.hashed_password == hashed_verify_password.hashed_password
    return result



import logging

from wiederverwendbar.logger.logger import Logger
from wiederverwendbar.logger.logger_settings import LoggerSettings
from wiederverwendbar.singleton import Singleton

LOGGER_SINGLETON_ORDER = 10


class LoggerSingleton(Logger, metaclass=Singleton, order=LOGGER_SINGLETON_ORDER):
    def __init__(self, name: str,
                 settings: LoggerSettings,
                 use_sub_logger: bool = True,
                 ignored_loggers_equal: list[str] = None,
                 ignored_loggers_like: list[str] = None):
        if ignored_loggers_equal is None:
            ignored_loggers_equal = []
        if ignored_loggers_like is None:
            ignored_loggers_like = []

        super().__init__(name, settings)

        self.ignored_loggers_equal = ignored_loggers_equal
        self.ignored_loggers_like = ignored_loggers_like

        if use_sub_logger:
            logging.setLoggerClass(SubLogger)

            for logger in logging.Logger.manager.loggerDict.values():
                if not isinstance(logger, logging.Logger):
                    continue
                _configure_logger(logger)


def _configure_logger(cls: logging.Logger):
    logger_singleton = LoggerSingleton()
    if cls.name in logger_singleton.ignored_loggers_equal or any([ignored in cls.name for ignored in logger_singleton.ignored_loggers_like]):
        return
    cls.setLevel(logger_singleton.level)
    cls.parent = logger_singleton


class SubLogger(logging.Logger):
    def __init__(self, name: str, level=logging.NOTSET):
        self.init = False
        super().__init__(name, level)
        _configure_logger(self)
        self.init = True

    def __setattr__(self, key, value):
        if key == "init":
            return super().__setattr__(key, value)
        if not self.init:
            return super().__setattr__(key, value)

    def setLevel(self, level):
        if not self.init:
            return super().setLevel(level)

    def addHandler(self, hdlr):
        if not self.init:
            return super().addHandler(hdlr)

    def removeHandler(self, hdlr):
        if not self.init:
            return super().removeHandler(hdlr)

    def addFilter(self, fltr):
        if not self.init:
            return super().addFilter(fltr)

    def removeFilter(self, fltr):
        if not self.init:
            return super().removeFilter(fltr)



import os
import tarfile
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Union, Optional


class TarRotatingFileHandler(RotatingFileHandler):
    """
    RotatingFileHandler that archives old log files as tar.gz files
    """

    def __init__(self,
                 name: str,
                 filename: Union[str, Path],
                 mode: str = "a",
                 max_bytes: int = 0,
                 backup_count: int = 0,
                 encoding: Optional[str] = None,
                 delay: bool = False,
                 archive_backup_count: int = 0):
        super().__init__(filename, mode, max_bytes, backup_count, encoding, delay)
        self.set_name(name)
        self.archiveBackupCount = archive_backup_count
        self.archiveBaseFilename = self.baseFilename[:self.baseFilename.rfind('.')]

    def doRollover(self):
        # rotate and delete old log files
        super().doRollover()

        # archive old log files if backupCount limit is reached
        if self.backupCount > 0:
            backup_log_pattern = self.baseFilename + '.%d'
            # get all existing backup logs
            backup_logs = [Path(backup_log_pattern % i) for i in range(1, self.backupCount + 1)]
            backup_logs = [log for log in backup_logs if log.exists()]

            # check if backup count limit is reached
            if len(backup_logs) >= self.backupCount:
                archive_filename_pattern = self.archiveBaseFilename + '_logs.%d.tar.gz'
                archive_count = 0
                # add all backup logs to tar archive
                archive_filename = archive_filename_pattern % archive_count
                while Path(archive_filename).exists():
                    archive_count += 1
                    archive_filename = archive_filename_pattern % archive_count

                    # if archive count limit is reached, delete oldest archive
                    if archive_count > self.archiveBackupCount:
                        archive_count = 0
                        archive_filename = archive_filename_pattern % archive_count
                        os.remove(archive_filename)
                        break

                with tarfile.open(archive_filename, 'w:gz') as tar:
                    for log in backup_logs:
                        tar.add(log, arcname=log.name)
                        os.remove(log)



import logging

from wiederverwendbar.logger.handlers.stream_console_handler import StreamConsoleHandler
from wiederverwendbar.logger.logger_settings import LoggerSettings

try:
    from wiederverwendbar.logger.handlers.rich_console_handler import RichConsoleHandler
except ModuleNotFoundError:
    RichConsoleHandler = None
from wiederverwendbar.logger.handlers.tar_rotating_file_handler import TarRotatingFileHandler


class Logger(logging.Logger):
    def __init__(self, name: str, settings: LoggerSettings):
        super().__init__(name, level=logging.DEBUG)

        self.settings = settings

        # add null handler
        null_handler = logging.NullHandler()
        self.addHandler(null_handler)

        # add console handler
        if self.settings.log_console:
            if RichConsoleHandler is None:
                ch = StreamConsoleHandler(
                    name=name,
                    console_outfile=self.settings.log_console_outfile
                )
            else:
                ch = RichConsoleHandler(
                    name=name,
                    console_outfile=self.settings.log_console_outfile,
                    console_width=self.settings.log_console_width,
                    show_time=self.settings.log_console_rich_show_time,
                    markup=self.settings.log_console_rich_markup,
                    show_level=self.settings.log_console_rich_show_level,
                    show_path=self.settings.log_console_rich_show_path
                )
            if self.settings.log_console_level is not None:
                ch.setLevel(self.settings.log_console_level.value)
            ch.setFormatter(logging.Formatter(self.settings.log_console_format))
            self.addHandler(ch)

        # add file handler
        if self.settings.log_file:
            # check if log_file_path is set
            if self.settings.log_file_path is None:
                raise ValueError("Log file path not set")

            # check if log_file_path parent directory exists
            if not self.settings.log_file_path.parent.exists():
                raise FileNotFoundError(f"Log file path parent directory not exist: '{self.settings.log_file_path.parent}'")

            fh = TarRotatingFileHandler(
                name=name,
                filename=self.settings.log_file_path,
                mode=self.settings.log_file_mode,
                max_bytes=self.settings.log_file_max_bytes,
                backup_count=self.settings.log_file_backup_count,
                encoding=self.settings.log_file_encoding,
                delay=self.settings.log_file_delay,
                archive_backup_count=self.settings.log_file_archive_backup_count
            )
            if self.settings.log_file_level is not None:
                fh.setLevel(self.settings.log_file_level.value)
            fh.setFormatter(logging.Formatter(self.settings.log_file_format))
            self.addHandler(fh)

        # log first message
        self.debug(f"Logger '{name}' initialized.")



import logging
from pathlib import Path
from typing import Optional

from wiederverwendbar.functions.eval_value import eval_value

logger = logging.getLogger(__name__)


def eval_file(src_file_path: Path, print_console: bool = False, **locals_vars) -> Optional[str]:
    """
    Evaluate file with variables.

    :param src_file_path Source file path
    :param print_console: Print evaluated value to console
    :param locals_vars: Variables to use
    :return: Evaluated file
    """

    logger.debug(f"Evaluate file '{src_file_path}'")

    # check if file exists
    if not src_file_path.is_file():
        logger.error(f"File '{src_file_path}' does not exist.")
        return None

    # read file
    with open(src_file_path, 'r') as f:
        src_file_raw = f.read()
        # split lines
        src_file_lines = src_file_raw.splitlines()

    # eval line by line
    evaluated_lines = []
    current_line = 1
    for line in src_file_lines:
        try:
            evaluated_line = eval_value(what=f"line '{current_line}' in file '{src_file_path}'",
                                        value=line,
                                        **locals_vars)
            evaluated_lines.append(evaluated_line)
        except Exception as e:
            if print_console:
                logger.error(f"Could not evaluate line {current_line} '{line}'. -> {e}")
            else:
                logger.debug(f"Could not evaluate line {current_line} '{line}'. -> {e}")
        current_line += 1

    # join lines
    evaluated = "\n".join(evaluated_lines)

    if print_console:
        logger.info(f"Evaluated file '{src_file_path}':\n{evaluated}")
    else:
        logger.debug(f"Evaluated file '{src_file_path}':\n{evaluated}")

    return evaluated



import logging
import os
from pathlib import Path
from typing import Any, Union, Optional

logger = logging.getLogger(__name__)


def eval_value(what: str,
               value: Union[str, Path, list[str]],
               index_marker: Optional[list[str]] = None,
               print_console: bool = False,
               use_environ: bool = True,
               **_local_vars) -> Union[str, Path, list[Union[str, Path]]]:
    """
    Evaluate value with variables.

    :param what: Message to log
    :param value: Value to evaluate
    :param index_marker: Index marker
    :param print_console: Print evaluated value to console
    :param use_environ: Use environment variables
    :param _local_vars: Variables to use
    :return: Evaluated value
    """

    # if index_marker is None, take default value
    if index_marker is None:
        index_marker = ["{{", "}}"]

    # check if len of index_marker is 2
    if len(index_marker) != 2:
        logger.error("Index marker must have a length of 2.")
        raise SystemExit(1)

    logger.debug(f"Evaluate {what}")

    def update_local_vars(_key: str, _value: Any) -> None:
        if _key in _local_vars:
            logger.error(f"Variable '{_key}' is reserved and cannot be used.")
            raise SystemExit(1)
        _local_vars[_key] = _value

    # add env to local_vars
    if use_environ:
        update_local_vars("env", dict(os.environ))

    as_list = True
    if not isinstance(value, list):
        value = [value]
        as_list = False

    def _format(_value: str) -> str:
        if type(_value) is not str:
            _value = str(_value)
        locals().update(_local_vars)
        if index_marker[0] in _value and index_marker[1] in _value:
            _start_index = _value.find(index_marker[0])
            _end_index = _value.find(index_marker[1])
            _expression = str(_value[_start_index + len(index_marker[0]):_end_index])
            _expression_value = eval(_expression)
            _expression_value_str = str(_expression_value)
            _solved_value = _value[:_start_index] + _expression_value_str + _value[_end_index + len(index_marker[1]):]
            return _format(_solved_value)
        else:
            return _value

    evaluated_values = []
    for v in value:
        try:
            if isinstance(v, Path):
                evaluated_v_path = Path()
                for part in v.parts:
                    evaluated_v = _format(part)
                    evaluated_v_path = evaluated_v_path / evaluated_v
                evaluated_v = evaluated_v_path
            else:
                evaluated_v = _format(v)
            evaluated_values.append(evaluated_v)
        except Exception as e:
            logger.error(f"Could not evaluate {what} '{v}'. -> {e}")
            raise SystemExit(1)

    if len(evaluated_values) == 0:
        logger.error(f"Could not evaluate {what}. -> Empty list")
        raise SystemExit(1)

    if not as_list:
        evaluated_values = evaluated_values[0]

    if print_console:
        logger.info(f"Evaluated {what} is: '{evaluated_values}'")
    else:
        logger.debug(f"Evaluated {what} is: '{evaluated_values}'")

    return evaluated_values



import logging
import sys

from wiederverwendbar.logger.logger_settings import LoggerSettings


def _resolve_file(outfile: LoggerSettings.TerminalOutFile):
    # choose file
    if outfile == LoggerSettings.TerminalOutFile.STDOUT:
        file = sys.stdout
    elif outfile == LoggerSettings.TerminalOutFile.STDERR:
        file = sys.stderr
    else:
        file = sys.stderr

    return file


class StreamConsoleHandler(logging.StreamHandler):
    def __init__(self, name: str, console_outfile: LoggerSettings.TerminalOutFile):
        super().__init__(stream=_resolve_file(console_outfile))
        self.set_name(name)
